"""Roustabout — structured markdown documentation of Docker environments."""

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("roustabout")
except PackageNotFoundError:
    __version__ = "0.0.0+dev"
